/*
**************************************************************
*                                                            *
*   Program Name :  extdata_vdif                             *
*                                                            *
*      Version:  1.11  2018-12-11                            *
*                                                            *
*  Copyright (c) 2018- T.Kondo All Right Reserved            *
*                                                            *
**************************************************************


  extract certain data from VDIF data.
  output data is usually without header block

          Ver. 2018-06-20 start development
               2018-07-23 update
               2018-12-11 debug

  How to run
      extdata_vdif vdif_file [options] 
    
         where  vdif_file -- vdif file name

                 options 
               -p[eriod] period
                     extracting period in sec (deflt 1.0)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0.0)
					 -VE : absolute value means skip over bytes 
               -k[ashima] [bytesframe] 
                     set header format to Kashima compulsory 
                     bytesframe: +VE bytes/frame set compulsory
			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion) (default)
									 2: Mark V          0,2,1,3
									 3: signed integer  1,0,2,3
               -thread n   -- set pick up thread number (default is 1st thread number)(2018-02-02)
               -fs fsMHz   -- set sampling frequency(MHz) for VDIF not including sampling fequency
               -o outfile  -- set output file name
               -band n     -- pick up band# (1-4) in case of CHs larger than 16

					
*/

#include "libvdif.h"

#define PROGRAM_VERSION "Ver 1.10  2018-12-11"

//#define HEAD_SEARCH_RANGE 600  // serach range (in 32bit word)  ==> move to libvdif.h
//#define VDIF_KASHIMA_SYNC 0xACABFEEDUL    // sync pattern for VDIF Kashima format ==> move to libvdif.h
#define BUFMAX 25000   // maximum read buffer size
#define DEFLT_VDIF_FILE "vdif.dat"
#define DEFLT_OUT_FILE "pickup_vdif.dat"

/*
typedef struct vdif_thread_info     // VDIF thread information type   added on 2018-02-05
{
    int nthread;    // # of threads
    int id[50];     // thread ID by thread
    int frame_count; //  continuous # of frame counter during unchanged thread ID by thread
} vdif_thread_info_t;
*/


typedef struct pro_runp     // run parameters for extdata_vdif
{
    char vdifname[256]; 
    char oname[256];
    double period;
	double soffset;
    int smode;
    int mode;
	int dmode;
	int on_sec_flag;
	int format;
    int legacy;
	int nmax;
	int all_flag; 

	int64_t skip_bytes;   // skip bytes (1st skip over # of bytes)
	int flag_compulsory;
    int e2bit;
    int thread_num;   // 2018-02-03
    double fsMHz;   // sampling frequency (MHz) 2018-02-07
    int flag_lsb_IQ;    // 2018-04-26 set LSB mode in case of I/Q data
    int flag_noout;
    int pickup_band;    // Band# 1-4 (in case of CHs included is larger than 16)
    int flag_m_thread_mode;   // 1: multi thread data mode 0: single thread data  2018-07-23
} pro_runp_t;

typedef struct pro_internal     // internal parameters
{
    char fname[256]; 
    double fbps;     // supposed data rate (bps)
	double fsample;  // supposed sampling frequency (Hz)
	double tframe;   // total number of frames (double) (computed from file size)
	double dlength;  // data length in seconds
    double pframe;   // frame length in sec (computed from headder info)

	int frames_per_sec;     // frames/sec per channel
	int badframes;   // number of invalid data frames
	int totalframes; // number of frames actually counted
    int adbit;
    int numch;
    double sfreq;
    int nfft;

    int flag_complex;
    int numthread;   // # of threads  2018-02-11

    int64_t fsize;   // file size (bytes)
    int flag_lsb_IQ;    // 2018-04-26 set LSB mode in case of I/Q data

    double period;
	double soffset;


} pro_internal_t;


void run_param_get(int argc, char *argv[], pro_runp_t *runp)                       
{
    char cbuf[256];
    char cbuf2[50];

    // initialize parameters
    runp->smode=0;
    runp->mode=0;
    runp->dmode=0;
	runp->nmax=0;
	runp->format=0;
	runp->on_sec_flag=0;
    runp->legacy=0;
	runp->all_flag=0;

    runp->vdifname[0]='\0';
	runp->period=1.0;
	runp->soffset=0.0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
    //runp->e2bit=2; // default
    runp->e2bit=1; // default  from 2016-08-03
    runp->thread_num=-1;
    runp->flag_lsb_IQ=0;
    runp->fsMHz=0.0;
    strcpy(runp->oname,DEFLT_OUT_FILE);
    runp->flag_noout=0;
    runp->pickup_band=-1;
    runp->flag_m_thread_mode=0; // 2018-07-23

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->vdifname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->vdifname[0]=='0'){  // set default name
          strcpy(runp->vdifname,DEFLT_VDIF_FILE);
	   }
    }

    /* command parameters get */
    if(cmnd_get(argc,argv,"-th",cbuf)==0){
        runp->thread_num = atoi(cbuf);    // thread #
        runp->flag_m_thread_mode=1;  // 2018-07-23
    }
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->period = atof(cbuf);     // period
	if(cmnd_get(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atof(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if(cmnd_get(argc,argv,"-k",cbuf)==0 ||cmnd_get(argc,argv,"-K",cbuf)==0 ){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
	if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode

    if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fsMHz = atof(cbuf);   // sampling frequency (MHz)

    if(cmnd_get(argc,argv,"-o",cbuf)==0) strcpy(runp->oname,cbuf);   // set out file
    if(cmnd_get(argc,argv,"-no",cbuf)==0) runp->flag_noout=1;   // set no output
    //if(cmnd_get(argc,argv,"-lsb",cbuf)==0) runp->flag_lsb_IQ=1;   // LSB mode in case of I/Q data
    if(cmnd_get(argc,argv,"-band",cbuf)==0) runp->pickup_band = atoi(cbuf);   // pickup band# 

}

void print_usage(void)
{
    printf("extdata_vdif %s\n",PROGRAM_VERSION);
    printf("\n");
    printf("Usage: extdata_vdif vdif_file [options]\n");
    printf("\n");
    printf("       where  vdif_file -- vdif file name\n");
    printf("                           if 0 is given, internal dflt name will be used\n");
    printf("\n");
    printf("          options \n");

    printf("               -p[eriod] period\n");
    printf("                     extracting period in sec (deflt 1.0)\n");
    printf("               -s[offset] soffset\n");
    printf("                     +VE : start offset (sec) (default is 0.0)\n");
    printf("                     -VE : absolute value means skip over bytes \n");
    printf("               -k[ashima] [bytesframe]\n");
    printf("                     set header format to Kashima compulsory \n");
	printf("                     bytesframe: +VE is bytes/frame set compulsory\n");
    printf("               -2bit mode -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)(default)\n");
    printf("                           2: Mark V standard 0,2,1,3\n");
    printf("                           3: signed integer  1,0,2,3\n");
    printf("               -th[read] n   -- set pick up thread number (default: 1st thread number)\n");
    printf("               -fs fsMHz   -- set sampling frequency(MHz) for VDIF not including sampling fequency\n");
    printf("               -o outfile  -- set output file name (deflt is %s)\n",DEFLT_OUT_FILE);
    printf("               -no[out]    -- suppress output file\n");
    printf("               -band n     -- pick up band# (1-4) in case of CHs larger than 16\n");
//    printf("               -lsb   -- set LSB mode in case of I/Q data (default is USB mode)\n");
    printf("\n");
}

void headerprint_for_debug(unsigned int *head, int headlen, vdif_header_t *vhead)
{
	int n;
	printf("\n");
	printf("========================= VDIF HEADER INFORMATION =========================\n");
	printf(" VDIF header (raw)        :");
	for(n=0; n<4; n++) printf(" %08X",head[n]); printf("\n");
	if(headlen > 4){
	printf("    extended header (raw) :");
	for(n=4; n<headlen; n++) printf(" %08X",head[n]); printf("\n");
	}
	printf("\n");
	printf(" Invalid flag = %d :", vhead->invalid);
	printf("       Legacy flag = %d\n", vhead->legacy);
	printf(" Sec from ref epoch = %d :", vhead->sec_fm_ref);
	printf("  Ref epoch = %d", vhead->ref_epo);
	printf("   (%04d/%02d/%02d %02d:%02d:%02d)\n", 
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss);

	printf(" Frame # = %d :", vhead->frame);
	printf("  VDIF Ver# = %d :", vhead->ver);
	printf(" # Chs(log2) = %d\n", vhead->log2_chs);
	printf(" Frame length in 8 byte unit = %d  (= %d bytes)\n", vhead->frame_len64,vhead->frame_len64*8);
	printf(" Data type = %d :", vhead->d_type);
	printf(" #bits/sample-1 = %d :", vhead->bpsampl);
	printf(" Thread ID = %d\n", vhead->thread_id);
	printf(" Station ID (A2) = %s ", vhead->sid);
	printf("   ( = %d in number)\n", vhead->station_id);
	printf(" EDV : %d\n", vhead->edv);
	printf(" Exteded Data W1 W2 W3 W4 : %06X %08X %08X %08X\n", vhead->w1,vhead->w2,vhead->w3,vhead->w4);

	printf(" --- in case of NICT Extended Format ---\n");	   
	//printf(" Head[4] = %08X :", vhead->w1);
	printf("  uflag = %d :", vhead->uflag);
	printf("  srate = %d :", vhead->srate);
	printf("  Sync block = %08X\n", vhead->sync_block);
	printf("  DAS/station name (A8) : %s\n", vhead->das);
	printf("===========================================================================\n");


}

void headerprint(vdif_header_t *vhead)
{
	printf("   Time        : %04d/%02d/%02d %02d:%02d:%02d  Frame# : %d\n", 
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss,
		   vhead->frame);
	printf("   Bytes/Frame : %d bytes\n", vhead->frame_len64*8);
	if(vhead->d_type==0){
	    printf("   Data type   : Real Data    ");
    } else {
	    printf("   Data type   : Complex Data ");
    }
    printf("        AD bits : %d", vhead->bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vhead->log2_chs));
	printf("   EDV#        : %d\n",vhead->edv);
	if(vhead->edv==1){   // This is NICT's EDV
       printf("   Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vhead->srate);
	   if(vhead->uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vhead->sync_block);
	   printf("      DAS/Station name   : %s\n", vhead->das);
	}
}


void disp_summary(pro_runp_t *runp, vdif_header_t *vhead1st, vdif_header_t *vheadlast,
				   pro_internal_t *ip)
{
/*
       summary display
*/

	printf("\n");
	printf("========================= SUMMARY =============================\n");
	printf(" File Name   : %s\n",runp->vdifname);
	printf("      Size   : %lld bytes\n",(long long)ip->fsize); 
    printf(" # of threads: %d\n",ip->numthread);  // 2018-02-11 
	printf(" Start Time  : %04d/%02d/%02d %02d:%02d:%02d\n", 
		   vhead1st->t.year,vhead1st->t.month,vhead1st->t.mday,vhead1st->t.hh,vhead1st->t.mm,vhead1st->t.ss);
	printf("               Thread ID : %d      Frame# : %d\n", vhead1st->thread_id,vhead1st->frame);
	printf(" Bytes/Frame : %d bytes\n", vhead1st->frame_len64*8);
	if(vhead1st->d_type==0){
	    printf(" Data type   : Real Data    ");
    } else {
	    printf(" Data type   : Complex Data ");
    }
    printf("    AD bits : %d", vhead1st->bpsampl+1);
    printf("    #Channels : %d\n", (int)pow(2.0, vhead1st->log2_chs));
	printf(" EDV#        : %d\n",vhead1st->edv);
	if(vhead1st->edv==1){   // This is NICT's EDV
       printf(" Extended header information (NICT\'s EDV)\n");
	   printf("      Sampling Frequency : %d", vhead1st->srate);
	   if(vhead1st->uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("      Sync block         : %08X\n", vhead1st->sync_block);
	   printf("      DAS/Station name   : %s\n", vhead1st->das);
	}
    printf(" # of frames/sec     : %d\n",ip->frames_per_sec);     
	printf(" # of total frames   : %d",ip->totalframes);     
	printf("   (#bad frames : %d)\n",ip->badframes);     
	//printf(" Data period (sec)   : %f\n",ip->dlength);     
    printf(" Data period (sec)/thread : %f\n",ip->dlength/(double)ip->numthread);  // 2018-02-11     
	printf(" Data Rate/thread (Mbps)  : %f\n",ip->fbps/1.0e6);
	//printf(" Sampling Frequency Estimated (MHz) : %f\n",ip->fsample/1.0e6); 
	if(vhead1st->d_type==0){
	   printf(" Sampling Frequency Estimated (MHz) for real : %f\n",ip->fsample/1.0e6);      // 2018-02-11
    } else {
	  printf(" Sampling Frequency Estimated (MHz) for complex : %f\n",0.5*ip->fsample/1.0e6);    // 2018-02-11 
    }
	printf("================================================================\n");

	printf("\n");

}


int get_thread_num(pro_runp_t *runp, int *numthread, int *table_thread, int *table_frames)
{
    // 2018-02-03

    int n;
    vdif_file_t vd;

    strcpy(vd.fname, runp->vdifname);

    n=vdif_thread_anal(&vd, numthread, table_thread, table_frames);

    return n;
}



int main(int argc, char *argv[])
{
    int nloop;
    int i,n,m, k, kaisu;
    int k_at_max;  // 2018-03-05
    double freq_at_max; // 2018-03-05
    int w;
    int nfft;
    int ymode; // amplitude scale mode   0: log   1: linear
    //int scount[16];  // allow max 16 channel
    int scount[64];  // allow max 16 channel 2018-02-03
	//int upos,ss; // for debug
	int r_thread_id;
    double fact;
    double *cdata;
    double *f_hanning;
    double theta;
    double sum,dc;
    float pows;
    float pmax,pmin,pmaxdbm,pmindbm,cofst;
    float ymaxdb,ymindb;
    double nfact;
    double tinteg;
    double telapsed;           // to measure processing time
    int flag_complex;  // 2018-02-09
    int flag_lsb;  // 2018-04-26
    int *idata, *qdata;  // 2018-02-09
    double *cspe;        // 2018-02-09
    int ndata;           // 2018-02-09
    int ndata2;          // 2018-03-01
    //int sphase;          // 2018-02-11
    int sphase[64];          // 2018-03-02
    int smphase;          // 2018-02-11
    int *rdata;           // 2018-02-11
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    char ltime1st[60];
    char pgdevice_dflt[20];
    char* dflt_fname=DEFLT_VDIF_FILE;
    pro_runp_t runp;
    pro_internal_t pt;
	//speanav_param_t st;
    vdif_file_t vd;
    vdif_header_t vhead1st;
    timeval_vdif_t time1st;
    int iwrite;
    int wbytes,twbytes;
    u_int32_t *wbuf;        // for data writing
    int flag_malloc_wbuf;
    int nband, band;
    FILE *fo;
    int *re_conv2bit;
    int re_conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)(default)
    int re_conv2bit_2[]={0,2,1,3};   //  conversion 2 bit
    int re_conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer


    if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){
          printf("extdata_vdif %s",PROGRAM_VERSION);
          print_usage();
          exit(0);
    } 
    if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){
          printf("extdata_vdif %s",PROGRAM_VERSION);
          exit(0);
    }

    run_param_get(argc, argv, &runp);

    if(runp.e2bit==1){
       re_conv2bit=re_conv2bit_1;
    } else if(runp.e2bit==3){
       re_conv2bit=re_conv2bit_3;
    } else {
       re_conv2bit=re_conv2bit_2;
    }

	strcpy(pt.fname,runp.vdifname);
	strcpy(vd.fname,runp.vdifname);
    pt.soffset=runp.soffset;
    pt.period=runp.period;



    pt.flag_lsb_IQ=runp.flag_lsb_IQ;


    printf("extdata_vdif %s\n",PROGRAM_VERSION);

    cstart1sec = clock();


	vd.skip_bytes=runp.skip_bytes;
	vd.flag_compulsory=runp.flag_compulsory;

    //if(vdif_t_ini(&vd,0)<0) goto exitend2;   // VDIF data check with auto mode
    if(vdif_t_ini(&vd,0)<0) exit(0);   // VDIF data check with auto mode
    //if(vdif_t_ini(&vd,1)<0) goto exitend2;  // VDIF data check with 1 sec reading compulsory
    if(vd.vhead.srate == 0){
        if(runp.fsMHz <= 0.0){
            printf(" This VDIF file does not include sampling frequency information!\n");
            printf(" Please set sampling frequency using \"-fsampl fmhz\" option.\n");
            exit(0);
        }
    }

	time1st=vd.vhead.t;  // save 1st time 

    if(runp.fsMHz != 0.0) vd.fsample=1.0e6*runp.fsMHz;  // 2018-02-07
    if(vd.fsample==0.0){
        if(runp.fsMHz==0.0){
            printf("Warning!  sampling frequency is not included in VDIF headder!!\n");
            printf("Please set sampling frequency using \"-fsampl fMHz\" option!!\n");
            exit(0);
        } else {
            vd.fsample=1.0e6*runp.fsMHz;
            vd.nframsec=(int)(1.0/((double)vd.frame_bytes*8.0/((double)(vd.adbit*vd.numch)*vd.fsample))+0.5); // 2018-12-11
        }
    } else {
        if(runp.fsMHz != 0.0){  // compulsory set sampling frequency  and ndramesec
            vd.fsample=1.0e6*runp.fsMHz;
            vd.nframsec=(int)(1.0/((double)vd.frame_bytes*8.0/((double)(vd.adbit*vd.numch)*vd.fsample))+0.5); 
        }
    }

    /* parameter set */
	pt.frames_per_sec=vd.nframsec;
	pt.adbit=vd.adbit;   // AD bits
	pt.sfreq=vd.fsample; 
	pt.numch=vd.numch;

    vd.bstream=vd.adbit*vd.numch;
    if(vd.vhead.d_type==1){  // complex data
        vd.bstream*=2;
    }

	
	/* 
	  set reading buffer size to satisfy following condition
	     1. (frame size)/(buffer size) should be an integer number
	*/
	vd.nword=vd.frame_bytes/4;     
	while (vd.nword > BUFMAX) vd.nword/=2;

	if(((vd.frame_bytes/4) % vd.nword) != 0){
		printf(" extdata_vdif: buffer size does\'nt satisfy the condition! %d %d\n",vd.nword, vd.frame_bytes*4);
	}

	vd.imax=(vd.frame_bytes/4)/vd.nword;   // # of loops to read entire one frame data
	vd.nsample=vd.nword*32/vd.adbit;  // # of samples in buffer (don't care about # of Chs)

	vd.imax_c=0;

//	vd.unitp=(double)vd.usampl/vd.fsample;   // period of usampl
    vd.framep=(double)(vd.frame_bytes*8/(vd.adbit*vd.numch))/vd.fsample;  // period of 1 frame


	//printf("imax nword32  %d %d\n",vd.imax,vd.nword);    //debug

    vd.ubuf  = (u_int32_t *)malloc(vd.nword*sizeof(u_int32_t)); //  data read buffer
    wbytes=vd.nword*sizeof(u_int32_t);

    nband=1;
    flag_malloc_wbuf=0;
    if(runp.pickup_band > 0 && vd.numch <= 16) runp.pickup_band=-1;
    if(vd.numch > 16 && runp.pickup_band > 0){ //   pick up 16 ch data
        //printf("debug %d\n", vd.bstream);
        if( (vd.bstream % (16*vd.adbit)) != 0){
            printf(" # of bits for 1 sample includes not integer # of bands!\n");
            exit(0);
        }
        nband=vd.bstream/(16*vd.adbit);

        if( (vd.nword % nband) != 0){
            printf(" read buffer word size is not multiple of # of bands!!\n");
            exit(0);
            //wbuf=(u_int32_t *)malloc(vd.nword*sizeof(u_int32_t));  // data write buffer
        } else {
            wbuf=(u_int32_t *)malloc((vd.nword/nband)*sizeof(u_int32_t));  // data write buffer
        }
        wbytes/=nband;
        flag_malloc_wbuf=1;
    } else {
        wbuf=vd.ubuf;
    }

	// file open
	if( (vd.fin  = fopen( vd.fname, "rb" )) == NULL ){
         printf("File (%s) Open Error! \n",vd.fname);
         goto exitend2;
    }
	vd.flag_open=1;



	if(vd.flag_vdif_anal==1){ // read pointer move to 1st valid header
		fseek(vd.fin,vd.pos_1st_header,SEEK_SET); // move pointer by K5->bytesec
	}

    if(runp.flag_m_thread_mode==1){     // 2018-07-23
        vd.pup_thread=runp.thread_num;
        printf(" Pick Up thread # ==> %d\n",vd.pup_thread); // 2018-02-03
    }

    printf(" 1st frame #  = %d\n",vd.vhead.frame);

    /* foward startoffset seconds */
	if(pt.soffset>0){
		printf("  %d second forwading.....",(int)pt.soffset);
        if(runp.flag_m_thread_mode==1){                          // 2018-07-23
	       if(fwdNsecVDIF_multi((int)pt.soffset, &vd) <0){       // 2018-07-23
              printf("Foward error %d\n",(int)pt.soffset);       // 2018-07-23
              i=fclose( vd.fin ) ;                               // 2018-07-23
              goto exitend;                                      // 2018-07-23
           }
        } else {
	       if(fwdNsecVDIF((int)pt.soffset, &vd) <0){
              printf("Foward error %d\n",(int)pt.soffset);
              i=fclose( vd.fin ) ;
              goto exitend;
           }
        }
	    printf("\n");
	    /* get headder info again */
        if(vdif_header_read_mode1(&vd)<0) goto exitend;
	    time1st=vd.vhead.t;
	}


	sprintf(ltime1st,"Time(UT) %04d/%02d/%02d %02d:%02d:%02d",
		time1st.year,time1st.month,time1st.mday,time1st.hh,time1st.mm,time1st.ss);
    //sprintf(ltime1st,"Debugd");  // debug

	// vdif read 

    if(runp.flag_m_thread_mode==0){     // 2018-07-23
       vd.pup_thread=vd.vhead.thread_id;  // set reading thread ID to 1st one  2018-02-09
       printf(" Pick Up thread # ==> %d\n",vd.pup_thread); 
    }

    //if(runp.thread_num >=0) r_thread_id=runp.thread_num;  // 2018-02-03
    //if(runp.thread_num >=0) vd.pup_thread=runp.thread_num;  // 2018-02-09

    //printf(" Pick Up thread # ==> %d\n",vd.pup_thread); // 2018-02-03

    if(runp.pickup_band > 0){
        printf(" Pick Up Band # ==> %d\n",runp.pickup_band); 
    }
    //r_thread_id=2;
        
    tinteg=0.0;
	vd.pos=0;

	//st.numch=1; // assuming 1 ch data

    nloop=0;

	vd.tot_frame=0; // total frame counter
	vd.upos=0;
	vd.usampl_c=0;
	vd.n_usampl=0;
	vd.flag_read_header=1;
	vd.flag_read_data=0;

    printf("  Reading thread ID is  %d\n",vd.pup_thread);
    printf("  Data reading.....\n");


    // out file open for write
    if(runp.flag_noout==0){
       if( (fo  = fopen( runp.oname, "wb" )) == NULL ){
           printf("File  %s  Open Error! \n",runp.oname);
           goto exitend;
       } 
       printf("Output File  %s  Opened! \n",runp.oname);
    }
    twbytes=0;

    //sphase=0;

loop1:

	    //upos=vd.upos;    // debug
		//ss=vd.vhead.t.ss; // debug

        i=vdif_read_frame(&vd, vd.pup_thread);
        if(nloop==0){  // save 1st header info
            vhead1st=vd.vhead;
        }

        pt.flag_complex=vd.vhead.d_type;    //   1: I/Q data 0: real data  2018-04-26
        flag_complex=vd.vhead.d_type;

        if(runp.pickup_band > 0){
            for(n=0; n< wbytes/(int)sizeof(u_int32_t); n++){        // supposing 1band = 4*bytes
                wbuf[n]=vd.ubuf[n*nband+runp.pickup_band-1];
            }
        }

        if(runp.flag_noout==0){
            //iwrite=fwrite(vd.head,sizeof(char),32,fo);  // header copy if set so (but usually no headder)
           //iwrite = fwrite( vd.ubuf, sizeof( char ), vd.frame_bytes, fo);
           iwrite = fwrite( wbuf, sizeof( char ), wbytes, fo); // data portion copy (write)
           twbytes+=iwrite;
        } else {
           twbytes+=wbytes;
        }


		tinteg+=vd.framep;
        if(tinteg >= pt.period) {
            printf("Start time   %4d/%03d %02d:%02d:%02d  Frame# %d Thread# %d\n",
                vhead1st.t.year,vhead1st.t.ddd,vhead1st.t.hh,vhead1st.t.mm,vhead1st.t.ss,vhead1st.frame,vhead1st.thread_id);

            printf(" Frame# , loop#  tinteg   %d  %d %f\n",vd.vhead.frame,nloop,tinteg);
            goto exitend;
        }
        if(vd.vhead.frame==0){
            printf("time   %4d/%03d %02d:%02d:%02d  Frame# %d Thread# %d\n",
                vd.vhead.t.year,vd.vhead.t.ddd,vd.vhead.t.hh,vd.vhead.t.mm,vd.vhead.t.ss,vd.vhead.frame,vd.vhead.thread_id);
        }
        nloop++;

        goto loop1;


  exitend: ;
    if(vd.flag_open==1){
		i=fclose( vd.fin ) ;
		vd.flag_open=0;
	}
      
    free(vd.ubuf);
    if(flag_malloc_wbuf==1) free(wbuf);

    if(runp.flag_noout==0){
       i=fclose(fo);
       printf("extdata_vdif:  created file is %s   (%d bytes)\n",runp.oname,twbytes);
    }
    //free(vd.idata);
 exitend2: ;

    return 0;
}

