/************************************************************************/
/*                                                                      */
/*   Program Name :  vdifsch2m                                          */
/*                                                                      */
/*     Vdifsch2m is VDIF (single-channel/thread) to                     */
/*                  VDIF (multi-channel/thread) converter               */
/*                                                                      */
/*                                                                      */
/*            Ver 1.20  2019-12-11                                      */
/*                                                                      */
/*      Copyright (c) 2019-     T.Kondo All Right Reserved              */
/*                                                                      */
/************************************************************************/
/*

     2019-10-23  development started
     2019-10-31  Version 1.0
     2019-11-06  revised and add "-chXX" and "-chall" options
     2019-11-10  revised
     2019-12-11  activate '-s -skip_bytes' option

   how to run
          vdifsch2m vdif_filename [options]
    
     where   vdif_filename -- VDIF data file name
              options (any order)

               -fsampl fmhz   -- channel sampling frequency (MHz) to compulsory set
			   -2bit mode     -- set 2bit AD decode table
			                         1: offset binary   0,1,2,3 (no conversion)
									 2: Mark V          0,2,1,3  (default for Mark5B)
									 3: signed integer  1,0,2,3

               -o out_name    -- VDIF file name to be created
                                 (default : see below)
               -d vdif_dir     -- directory for VDIF file out
                                 (default  : vdif file directory)
               -s[offset] soffset
			         +VE : start offset (sec) (default is 0)
					 -VE : absolute value means skip over bytes  (default 0)
               -skip skip_bytes   --- set 1st skip_bytes  (same as -s -VE)
               -k[ashima] [bytesframe] 
                     set header format to Kashima compulsory 
                     bytesframe: +VE bytes/frame set compulsory
               -type0   --- naming type 0  vdifname + '.mch'
               -type1   --- naming type 1  (add .mch before file extention

               -sid thread_id  -- set thread# in case of output VDIF (default is 0)
               -sync[1sec]   -- adjust data start time on the second
               -ch1 nn     -- set VDIF thread number corresponding to CH#1
                              nn is VDIF thread number
               -ch2 nn     -- set VDIF thread number corresponding to CH#2
                      ...........................
               -chXX nn    -- set VDIF thread number corresponding to CH#XX
                               XX is CH# (1-32)
               -chall n1:n2:....:nXX
                           -- set VDIF thread numbers corresponding to all channels
                               n1  -- VDIF thread number corresponding to CH#1
                               n2  -- VDIF thread number corresponding to CH#2
                                ...............
                               nXX -- VDIF thread number corresponding to CH#XX
                                      XX is 1-32
 
  Naming rule for VDIF file created (VDIF.vdif : original VDIF name)
            Type 0:  VDIF.vdif.mch 
            Type 1:  VDIF.mch.vdif
*/

#define PROGRAM_VERSION "vdifsch2m (Ver 1.20  2019-12-11)"

#include "vdifsch2m.h"

#define DFLT_DATA_FILE "vdif.dat"
#define BUFMAX 25000   // maximum read buffer size
#define DFLT_FRAME_SIZE 2000 // out VDIF frame size (except for header)

int cmnd_get_num(int argc, char *argv[], char *command, char *param, int *num)
// get argument like -command parameter
// search "-command" in argument, if it is exsted, 0 is returned as
// a function value, and next argument is returned as a parameter (ascii).
// if "-command" is not exsited, -1 is returned as a function value.
// if "-command" is not accompanied by real parameter, NULL character
// is returned as param.
// 
{
    int n;
    char cbuf[80];

    param[0]='\0';

    //search command
    //printf("%s\n",command); // debug
    n=2;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,command,strlen(command))==0){  
            n++;
			*num=n;
            if(n > argc) return 0;
            strcpy(param,*(argv+n-1));
			*num++;
            return 0;
        }
        n++;
    }
    return -1;
}

int vex_parse(char *cbuf, vparam_t *vparam)  
{
/*
    Parse VEX parameters by '=',':',';', and ' '.
    '=',':',';' are returned as a parsed item for further analysis
    Total number of items including '=',':',';' is defined in
    type define statesment of vparam as

    typedef struct vparam
    {
        int  n;          //number of param
        char param[60][128];   // 60x128 characters
    } vparam_t;

    In this case 60 is a maximum number of items and
    128 characters are maximum char in an item.

    If runtime error occurrs, please try to increase the number of
    items in the vparam definition.


*/

    // vex_parse =0;  single statement in a cbuf
    //           =1;  more statement left in a cbuf, start point in cbuf 
    //                is updated 
    char abuf[500];
    char *cset=" ";
    char *p=NULL;
    int i,ipos,k,kpos,flag,mflag,len;

    len=strlen(cbuf);
    //printf("len = %d\n",len);

    //remove extra space in front

    if(cbuf[0]==' '){
        kpos=0;
        for(k=0; k<len; k++){
            if(cbuf[k]!=' '){
                kpos=k;
                break;
            }
        }
        i=0;
        for(k=kpos; k<len; k++){
            cbuf[i]=cbuf[k];
            i++;
        }
        cbuf[i]='\0';
    }
    // len again

    len=strlen(cbuf);


    // parse start

    if(cbuf[0] == '*') { // comment
        vparam->n=0; 
        return 0;
    }
    i=0;
    ipos=0;
    flag=0;
    for(k=0; k<len; k++){
        if(cbuf[k] == ' ' || cbuf[k] == '=' || cbuf[k] == ':'){
            if(flag==1) {  // one parameter end
                vparam->param[ipos][i]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            if(cbuf[k] != ' '){
                vparam->param[ipos][i]=cbuf[k];
                vparam->param[ipos][i+1]='\0';
                flag=0;
                i=0;
                ipos++;
            }
            goto nextk;
        }
        if(cbuf[k] == ';'){
            kpos=k;
            goto exithere;
        }
        if(cbuf[k] != ' '){
            flag=1;
            vparam->param[ipos][i]=cbuf[k];
            i++;
        }
nextk:;
    }
    kpos=k;
exithere:
    vparam->param[ipos][i]='\0';
    ipos++;
    vparam->n=ipos;
    i=0;
    mflag=0;
    if(kpos < len-1){
        for(k=kpos+1; k<len; k++) {
            if(cbuf[k] != '\0'){
                abuf[i]=cbuf[k];
                i++;
            }
            if(cbuf[k] == ';') mflag=1;
        }
    }
    if(mflag==1){
        abuf[i]='\0';
        strcpy(cbuf,abuf);
        return 1;
    }
    return 0;
}


void run_param_get(int argc, char *argv[],runp_t *runp)                       
{
    int n,i;
    char cbuf[256];
    char key[10];
    vparam_t vp;

	// initialize parameters

    /* common parameters */
    runp->ptype=1;
    runp->vdifname[0]='\0';
    //runp->k5name[0]='\0';
    runp->oname[0]='\0';
    runp->oname_given[0]='\0';
    runp->soffset=0;
    runp->period=0;
    runp->odd=0;

    /* mode 1 parameters */
    runp->channel=0;
    runp->group=0;
    runp->unit=0;
    runp->reverse=0;
    runp->info_file[0]='\0';
    runp->outdir[0]='\0';
    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    runp->monit=0;
	runp->flag_compulsory=0;
	runp->skip_bytes=0;
	runp->k5numch=1; // 
    runp->vdifch=1;
    runp->naming_type=0;
    runp->flag_ch_alloc=0;
    runp->pickup_tid=-1;
    runp->flag_1sec_adjust = 0;

    /* mode 2 parameters */
    runp->numch=0;
    runp->adbit =0;
    //runp->fmhz  =(double)M5SAMPLE;
    runp->fmhz  =0.0;
	runp->nbitstream=0;
    for(n=0; n<32; n++) runp->ibit[n]=0;
    runp->fmode=3;   // default conversion file mode 1: 1ch  2: 4ch  3: 4x4ch  4: 2x4ch
    //for(n=0; n<4; n++) runp->chtable[n]=0;
    for(n=0; n<32; n++) runp->chtable[n]=n;  // default thread# for channel
    runp->e2bit=1;     // VDIF default
    runp->set_tid = 0;  // default setup thread# 

    //runp->k5nfile=0; //  0 means auto mode  2016-08-08

    /* 1st parameter (vdif file name) get */
    if (argc>=2){
       strcpy(runp->vdifname,*(argv+1));
	   if (strlen(*(argv+1))==1 && runp->vdifname[0]=='0'){  // set default name
          strcpy(runp->vdifname,DFLT_DATA_FILE);
	   }
    }
     /* check VDIFfile name is acompanied by directory or not */
    //get_true_path(runp->vdifname, M5DIR, "./",cbuf);
    //strcpy(runp->vdifname,cbuf);

    /* command parameters get */
    if(cmnd_get_p(argc,argv,"-o",cbuf)==0) strcpy(runp->oname_given,cbuf);  // output file name
	if(cmnd_get_p(argc,argv,"-d",cbuf)==0){              // out dir
		strcpy(runp->outdir,cbuf);  
        /* env variable is substituted by real one */
        //if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
        //   strcpy(runp->outdir,cbuf);
        //}
	}
//    if(cmnd_get(argc,argv,"-c",cbuf)==0) runp->channel = atoi(cbuf);  // CH
//    if(cmnd_get(argc,argv,"-g",cbuf)==0) runp->group   = atoi(cbuf);  // group
    //if(cmnd_get_p(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);  // start offset
    if(cmnd_get_p(argc,argv,"-p",cbuf)==0) runp->period  = atoi(cbuf);  // period
	if(cmnd_get_p(argc,argv,"-r",cbuf)==0) runp->reverse=1;  // bit reverse mode
	//if(cmnd_get_p(argc,argv,"-si",cbuf)==0) strcpy(runp->sid,cbuf);  // station id
	if(cmnd_get_p(argc,argv,"-scan",cbuf)==0) {  // scan
		  runp->scan=atoi(cbuf);
          if(runp->scan < 1) runp->scan=1;
	}
	if(cmnd_get_p(argc,argv,"-monit",cbuf)==0) runp->monit=1;  // monitor mode on
    //if(cmnd_get_p(argc,argv,"-fs",cbuf)==0) runp->fmhz  = atof(cbuf);  // VDIF sampling frequency
    if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fmhz  = atof(cbuf);  // VDIF sampling frequency  2018-12-11
    if(cmnd_get_p(argc,argv,"-ad",cbuf)==0) runp->adbit = atoi(cbuf);  // VDIF AD bits
    if(cmnd_get_p(argc,argv,"-bs",cbuf)==0) runp->nbitstream = atoi(cbuf);  // bit stream

    if(cmnd_get(argc,argv,"-2bit",cbuf)==0) runp->e2bit = atoi(cbuf);   // 2bit encode mode
	if(cmnd_get_p(argc,argv,"-s",cbuf)==0){
		if(atof(cbuf) >= 0.0){
		    runp->soffset = atoi(cbuf);   // second offset
		} else {
			runp->skip_bytes=-atoi(cbuf);  // skip bytes
		}
	}
    if (cmnd_get_p(argc, argv, "-soffset", cbuf) == 0) {
        if (atof(cbuf) >= 0.0) {
            runp->soffset = atoi(cbuf);   // second offset
        }
        else {
            runp->skip_bytes = -atoi(cbuf);  // skip bytes
        }
    }
    if (cmnd_get_p(argc, argv, "-skip", cbuf) == 0) {  // 2019-12-11
        runp->skip_bytes = atoi(cbuf);  // skip bytes
    }
    if(cmnd_get_p(argc,argv,"-k",cbuf)==0){ // set compulsory mode
		runp->flag_compulsory=1;
        if(cbuf[0] != '\0'){
            if(cbuf[0] != '-') runp->flag_compulsory=atoi(cbuf);
        }
    }
    if(cmnd_get_p(argc,argv,"-type1",cbuf)==0) runp->naming_type=1;
    //if(cmnd_get_p(argc,argv,"-pid",cbuf)==0) runp->pickup_tid = atoi(cbuf);   // pickup thread ID
    if (cmnd_get_p(argc, argv, "-sid", cbuf) == 0) runp->set_tid = atoi(cbuf);   // setup thread ID
    if(cmnd_get_p(argc, argv, "-sync", cbuf) == 0) runp->flag_1sec_adjust = 1;
                  // adjust data start on the second
    
    for (n = 0; n < 32; n++) {
        if (n < 10) {
            sprintf(key, "-ch%1d", n + 1);
            if (cmnd_get_p(argc, argv, key, cbuf) == 0) {  // CHX set
                runp->chtable[n] = atoi(cbuf);
                runp->flag_ch_alloc = 1;
            }
            sprintf(key, "-ch%02d", n + 1);
            if (cmnd_get_p(argc, argv, key, cbuf) == 0) { // CH0X set
                runp->chtable[n] = atoi(cbuf);
                runp->flag_ch_alloc = 1;
            }
        }
        else {
            sprintf(key, "-ch%2d", n + 1);
            if (cmnd_get_p(argc, argv, key, cbuf) == 0) {  // CHXX set
                runp->chtable[n] = atoi(cbuf);
                runp->flag_ch_alloc = 1;
            }
        }
    }


    if(cmnd_get_n(argc,argv,"-chall",cbuf)==0) { // all channel set
        vex_parse(cbuf, &vp);
        i=0;
        for(n=0; n<vp.n; n++) {
           //printf(" n= %d  %s\n",n,vp.param[n]);// debug
            if(vp.param[n][0]!=':'){
                runp->chtable[i]=atoi(vp.param[n]);
                i++;
            }
        }
        runp->flag_ch_alloc = 1;

    }

}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("    vdifsch2m vdif_file [options]\n");
    printf("\n");
    printf("where   vdif_file  -- VDIF data file name\n");
    printf("      options (any order)\n");
    printf("        -o oname     -- VDIF file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d vdifdir      -- directory for output VDIF file\n");
    printf("                          (default  : same as VDIF file directory)\n");
    //printf("        -bs nbitsream -- set Mark5B # of bit-stream (1,2,4,8,16,32)\n");
    printf("        -fsampl fmhz  -- channel sampling frequency (MHz) to compulsory set\n");
//    printf("        -u unit       -- pickup unit# (1-4) for 4 ch conversion mode\n");
//    printf("                         (-1ch option is stronger than this option)\n");
//    printf("                         (default 4x4ch mode)\n");
//    printf("        -1ch vdifch   -- set K5 data to 1ch mode and \n");
//    printf("       	                    pick up VDIF ch (starting at 1)\n");
//    printf("        -4ch ch1 ch2 ch3 ch4\n");
//    printf("                      -- set K5 data to 4 ch mode and  \n");
//    printf(" 	                       pick up VDIF 4 channels\n");
    printf("        -2bit mode    -- set 2bit AD decode table\n");
    printf("                           1: offset binary   0,1,2,3 (no conversion)\n");
    printf("                           2: Mark V standard 0,2,1,3 (default)\n");
    printf("                           3: signed integer  1,0,2,3\n");
//    printf("        -s[offset] soffset\n");
//    printf("                      -- set start offset (sec) (default is 0.0)\n");
    printf("        -skip skip_bytes\n");
    printf("                      -- set 1st skip over bytes (default 0) \n");
//    printf("        -s strat_sec  -- offset time from data head (sec)\n");
//    printf("                          (default :  0 )\n");
//    printf("        -sync[1sec]   -- adjust data start time on the second\n");
    printf("        -p period     -- data period to convert (sec)\n");
//    printf("                          (default : all)\n");
//    printf("        -r            -- reverse track order when fanout =2 or more\n");
//    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
//    printf("                              file creation mode\" is set\n");
//    printf("        -sid stat_id  -- station ID when \"information file creation\n");
//    printf("                              mode\" is set\n");
//    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
//    printf("                              creation mode\" is set (default : 1)\n");
//    printf("        -subp subpass -- set subpass at information file\n");
//    printf("                              creation mode (default is \"A\")\n");
    printf("        -type1        -- outfile naming Type 1 (default Type 0)\n");
    //printf("        -pid thread_id -- set pickup thread ID compulsory\n");
    printf("        -sid thread_num -- set output thread#  (default 0)\n");
    printf("        -ch1 nn     -- set VDIF thread number corresponding to CH#1\n");
    printf("                          nn is VDIF thread number\n"); 
    printf("        -ch2 nn     -- set VDIF thread number corresponding to CH#2\n");
    printf("           ...........................\n"); 
    printf("        -chXX nn    -- set VDIF thread number corresponding to CH#XX\n"); 
    printf("                          XX is CH# (1-32)\n"); 
    printf("        -chall n1:n2:....:nXX\n");
    printf("                 -- set VDIF thread numbers corresponding to all channels\n"); 
    printf("                     n1  -- VDIF thread number corresponding to CH#1\n"); 
    printf("                     n2  -- VDIF thread number corresponding to CH#2\n"); 
    printf("                      ...............\n"); 
    printf("                     nXX -- VDIF thread number corresponding to CH#XX\n");
    printf("                          XX is 1-32\n"); 
    printf("\n");

    printf("   Naming rule for VDIF file created (VDIF.vdif : original VDIF name)\n");
    printf("        Type 0  VDIF.vdif.mch (default)\n");
    printf("        Type 1  VDIF.mch.vdif\n");

    printf("\n");
}

void monit_runp(runp_t *runp)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ---------\n");
       printf("   vdifname(in)  : %s\n",runp->vdifname);
       printf("   vdifname(out) : %s\n",runp->oname_given);
       printf("   outdir    : %s\n",runp->outdir);
       printf("   bitstream : %d\n",runp->nbitstream);
       printf("   fmhz      : %f\n",runp->fmhz);
       printf("   adbit     : %d\n",runp->adbit);
       printf("   skip_bytes: %lld\n",(long long)runp->skip_bytes);
//       printf("   k5numch   : %d\n",runp->k5numch);
//       printf("   vdif_channels : %d %d %d %d\n",runp->chtable[0],
//		   runp->chtable[1],runp->chtable[2],runp->chtable[3]);
//       printf("   2bit mode : %d\n",runp->e2bit);
//       printf("   infofile  : %s\n",runp->info_file);
//       printf("   channel   : %d\n",runp->channel);
//       printf("   group     : %d\n",runp->group);
//       printf("   soffset   : %d\n",runp->soffset);
//       printf("   period    : %d\n",runp->period);
//       printf("   odd       : %d\n",runp->odd);
//       if(runp->vex_file[0] != '\0')
//          printf("   vex_file  : %s\n",runp->vex_file);
//       if(runp->sid[0] != '\0' && runp->sid[0] !='-')
//          printf("   station_id: %s\n",runp->sid);
       printf(" -------------------------------------\n");
    printf("\n");
}

int main(int argc, char *argv[])
{
       int i;
       double telapsed;           // to measure processing time
       char fname[256];
       char program_version[60];
	   time_t tstart, tfinish;    
       runp_t runp;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       }
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 

       strcpy(program_version,PROGRAM_VERSION);

       /* run parameters get */
       run_param_get(argc, argv, &runp); 

       /* out directory create if not set */

       if(runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.vdifname, runp.outdir, fname); // debug
          //printf("A %s\n",runp.mk5name);
          //printf("A %s\n",runp.outdir);
          //printf("A %s\n",fname);
       }


       /* Program Name Display */
       printf("*****************************************************************\n");
       printf("*    VDIF(single-ch/thread) to VDIF(multi-ch/thread) Converter  *\n");
       printf("*        %32s  by  T.KONDO          *\n",program_version);
       printf("*****************************************************************\n");
       printf("\n");

//       monit_runp(&runp);

       tstart= time(NULL);


       i=convVDIF2VDIF2(&runp);

       if(i < 0) goto exitend;

       //printf("  K5 file (%s) has been created.\n\n",runp.k5name);


       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("\n Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
       return 0;
}


/* UTC to MJD conversion */
int ut2mjd(double *d, int year, int month, int day,
           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
  [return]
     ut2mjs -- MJD  (integer day)
     d      -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;

	mnth=month;
	y4=year;
    bc = year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = year > 1582 || (year == 1582 && month > 10) ||
              (year == 1582 && month == 10 && day >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (hour < 12) {
        j = 0;
        *d = 0.5;
    } else {
        j = 1;
        *d = -0.5;
    }
    *d += (hour*3600L + min*60 + sec)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
    j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + day;
    
	/* convert to MJD */
	if (hour < 12) {
		*d -= 0.5;
	} else {
		*d += 0.5;
		j -= 1L;
	}
    j -= 2400000L;
    return j;
}

void ch_realloc(M5B_HEADER_T *m5b, int *chtable)
{
// supposed maximum bitstream 2bit x 16ch
//  Mark5B is 1bit or 2bit AD

    int i,n,m,numch,adbit,bs;
    int pch;
    unsigned int iw, iw32, ow32;
    
    unsigned int imask[4]={0x1, 0x3, 0xF, 0xFF};
    unsigned int umask[4]={0x7FFFFFFF, 0x3FFFFFFF, 0x0FFFFFFF, 0x00FFFFFF};

    adbit=m5b->si.adbit;
    numch=m5b->si.numch;
    bs=adbit*numch;

    // convert each 32-bit word
    for(i=0; i<m5b->ndata; i++){
        iw32=m5b->idata[i];
        ow32=0;
        for(n=0; n< 32/bs; n++){
            for(m=0; m<numch; m++){
                ow32 >>= adbit;
                pch=chtable[m];
                if(pch > numch) pch=m+1;  // doesn't change
                iw= iw32 >> adbit*(pch-1);
                iw &= imask[adbit-1];
                iw <<= 32-adbit;
                ow32= (ow32 & umask[adbit-1]) | iw;
            }
            iw32 >>= bs;
        }
        m5b->idata[i]=ow32;
    }
}



void disp_condition2(runp_t *runp, vdif_file_t *vd, int numthread, int *table_thread)
{
/*
       display VDIF and conversion infomation
*/

    int n;
	printf("\n");
	printf(" ============================ VDIF FILE INFO ==================================\n");
    printf("  File Name   : %s\n",runp->vdifname);
    printf("       Size   : %lld bytes\n",(long long)vd[0].f_size); 
	printf("  Start Time  : %04d/%02d/%02d %02d:%02d:%02d\n", 
        vd[0].vhead.t.year,vd[0].vhead.t.month,vd[0].vhead.t.mday,vd[0].vhead.t.hh,vd[0].vhead.t.mm,vd[0].vhead.t.ss);
    printf("                Thread ID : %d      Frame# : %d\n", vd[0].vhead.thread_id,vd[0].vhead.frame);
    printf("  # of Threads: %d\n",numthread);
    printf("  Thread #    : ");
    for (n = 0; n < numthread; n++) printf(" %2d", table_thread[n]); printf("\n");

    //if(runp->pickup_tid >0){
    //    printf("%d\n",runp->pickup_tid);
    //} else {
    //    printf("%d\n",vd->vhead.thread_id);
    //}
	printf("  Bytes/Frame : %d bytes\n", vd[0].vhead.frame_len64*8);
	if(vd[0].vhead.d_type==0){
	    printf("  Data type   : Real Data    ");
    } else {
	    printf("  Data type   : Complex Data ");
    }
    printf("     AD bits : %d", vd[0].vhead.bpsampl+1);
    printf("     #Channels : %d\n", (int)pow(2.0, vd[0].vhead.log2_chs));
	printf("  EDV#        : %d\n",vd[0].vhead.edv);
	if(vd[0].vhead.edv==1){   // This is NICT's EDV
       printf("  Extended header information (NICT\'s EDV)\n");
	   printf("       Sampling Frequency : %d", vd[0].vhead.srate);
       if(vd[0].vhead.uflag==0){
		   printf(" kHz\n");
	   } else {
		   printf(" MHz\n");
	   }
	   printf("       Sync block         : %08X\n", vd[0].vhead.sync_block);
	   printf("       DAS/Station name   : %s\n", vd[0].vhead.das);
	}
//	printf(" # of frames/sec     : %d\n",ip->framsec);     
//	printf(" # of total frames   : %d",ip->totalframes);     
//	printf("   (#bad frames : %d)\n",ip->badframes);     
//	printf(" Data period (sec)   : %f\n",ip->dlength);     
//	printf(" Data Rate/thread (Mbps) : %f\n",ip->fbps/1.0e6);     
//	printf("  Sampling Frequency Estimated (MHz) : %f\n",vd[0].fsample/1.0e6); 
//	printf("  Sampling Frequency Estimated (MHz) : %f\n",runp->fsample_orig/1.0e6); 
    if (runp->fmhz > 0.0) {
        printf("  Sampling Frequency set by Option (MHz) : %f\n", runp->fmhz);
        if (runp->fmhz != runp->fsample_orig/1.0e6) {
            printf("  Warning!!  Sampling frequency set by option (%.1f MHz) is different from\n",
                runp->fmhz);
            printf("             that estimated from VDIF data (%.1f MHz)\n",
                runp->fsample_orig/1.0e6);
            printf("  This causes an extra delay offset if VDIF data are not started on the second!!\n");

        }
    }
    else {
        printf("  Sampling Frequency Estimated (MHz) : %f\n", vd[0].fsample / 1.0e6);
    }
  

    printf(" ==============================================================================\n");

	printf("\n");

}

void disp_condition_ch_thread(runp_t *runp, vdif_file_t *vd, int numthread, int *table_thread)
{
    /*
           display VDIF and conversion infomation
           CH# and Thread# relations
           2019-11-06
    */

    int n;
    printf(" ======================= Relation Between CH# and Thread# =====================\n");
    printf("   CH#      : ");
    for (n = 0; n < numthread; n++) printf(" %2d", n + 1); printf("\n");
    printf("   Thread#  : ");
    for (n = 0; n < numthread; n++) printf(" %2d", table_thread[n]); printf("\n");
    printf(" ==============================================================================\n");
    printf("\n");
}



int fwdNextSecondDIF(vdif_file_t *vd)
{
/* 
       forward next on the second with in the same thread ID
	   if current position is not on the second,
	   1sec forward means forward at the begining of the next second
	   (so less than 1 sec forward)
*/
	int64_t pos,posk;
    //int r_thread_id;


    if(vd->flag_open==0){                                     // 2018-12-12
	    /* VDIF file open for conversion */
	    if( (vd->fin  = fopen( vd->fname, "rb" )) == NULL ){    // 2018-12-12
             printf("File (%s) Open Error! \n",vd->fname);     // 2018-12-12
             return -1;                                       // 2018-12-12 
        }                                                     // 2018-12-12
	    vd->flag_open=1;                                      // 2018-12-12
    }                                                         // 2018-12-12


    pos=vd->pos;  // save current file pointer
    //r_thread_id=vd->vhead.thread_id;

	// compute set position

    //posk=pos+vd->vhead.frame_len64;
    posk=pos+vd->vhead.frame_len64*8;  // 2018-12-12
    
	while(posk < vd->f_size){  // 
		fseek(vd->fin,posk,SEEK_SET); // move pointer by K5->bytesec
        if( ferror( vd->fin ) ){
            return -1;
        }
		vd->pos=posk;
	    /* get headder info */
        if(vdif_header_read_mode1(vd)<0) return -1;
        //if(vd->vhead.frame == 0 && vd->vhead.thread_id==r_thread_id) goto exit2return;
        if(vd->vhead.frame == 0 && vd->vhead.thread_id==vd->pup_thread) goto exit2return;
        posk+=vd->vhead.frame_len64*8; // 2018-12-12
    }
exit2return:       
    return 0;
}


void get_frame_num(double tf1, int nf1, double tf2, int *nf11, int *nf2)
{
/*
     get frame # for VDIF out
     2019-10-23

     [in]
         tf1  -- frame period of VDIF in
         nf1  -- frame# of VDIF in
         tf2  -- frame period of VDIF out
     [return]
         nf11 -- frame# of VDIF in to satisfy integer frame# in VDIF out
         nf2  -- frame# of VDIF out when frame# of VDIF in is nf11

*/
    double res;
    double ratio;
    int nw1,nw2;
    int nmax;
    nmax = (int)(1.0 / tf1 + 0.5);
    nw1 = 0;
    nw2 = 0;
    ratio = tf2 / tf1;
    if (nf1 > 0) {
        nw1 = nf1;
        while (1) {
            res = fmod((double)nw1, ratio);
            if (fabs(res) <= 1.0e-6) {
                nw2 = (int)((double)nw1 / ratio + 0.5);
                break;
            }
            nw1++;
            if (nw1 > nmax) {
                nw1 = 0;
                break;
            }
        }
    }
    *nf11 = nw1;
    *nf2 = nw2;
}

//int get_num_thread_vdif(char *vdif_name, int *numthread, int *table_thread, int *table_frames)
//int get_num_thread_vdif(char *vdif_name, vdif_thread_info_t *thd)
int get_num_thread_vdif(char *vdif_name, int64_t skip_bytes, vdif_thread_info_t *thd)
{
/*
     get # of thread in VDIF file
      2019-10-23
      2019-12-11 add skip_bytes in arguments

   [in]
       vdif_name   --- VDIF file name
       skip_bytes  --- 1st skip bytes in VDIF file
   [out]
       numthread   --  # of threads
       tab_thread[ ] -- thread# table
       tab_frames[ ] -- # of frames during thread ID doesn't change  (# of frames per second for the case of numthread=1)

   [return value]
           +VE:   # of threads
            -1: not a VDIF format
           -99: file open error or file read error
*/
    int iret;
    vdif_file_t *vd;
    vd = (vdif_file_t *)malloc(sizeof(vdif_file_t));
    strcpy(vd->fname, vdif_name);
    vd->skip_bytes = skip_bytes;  // 2019-12-11
    iret = vdif_thread_anal(vd, &thd->numthread, thd->tab_thread, thd->tab_frames);
    free(vd); // 2019-12-11
    return iret;
}

int get_vdif_basic_info(vdif_file_t *vd, runp_t *runp)
{
    /*
        get VDIF basic parameter
        VDIF file is closed at end
        2019-10-24

    */
    int ierr;
    ierr = 0;
    strcpy(vd->fname, runp->vdifname);
    vd->skip_bytes = runp->skip_bytes;
    vd->flag_compulsory = runp->flag_compulsory;
    vd->soffset = runp->soffset;
    /* get VDIF file info by opening and get reading file (then close) */
    if (vdif_t_ini(vd, 0) < 0) {
        //if (vdif_t_ini(&vd, 1) < 0) {// VDIF data check with 1 sec reading compulsory
        ierr = -1;
        goto exit2return;   // VDIF data check with auto mode
    }
    runp->fsample_orig = vd->fsample;  // keep fsample got from VDIF data 2019-11-10
    if (vd->fsample == 0.0) {
        if (runp->fmhz == 0.0) {
            printf("get_vdif_basic_info: Warning!  sampling frequency is not included in VDIF headder!!\n");
            printf("get_vdif_basic_info: Please set sampling frequency using \"-fsampl fMHz\" option!!\n");
            exit(0);
        }
        else {
            vd->fsample = 1.0e6*runp->fmhz;
        }
    }
    else {
        if (runp->fmhz != 0.0) {  // compulsory set sampling frequency
            vd->fsample = 1.0e6*runp->fmhz;
        }
    }
exit2return:
    return ierr;
}

int ini_vdif_set_1st_frame(vdif_file_t *vd, int thread_num)
{
    /*
        set up VDIF parameter and set pointer to 1st frame of given thread#
        
        2019-10-24

    */
    int ierr;
    ierr = 0;

    /* pick up thread # set */
    vd->pup_thread = thread_num;  // set reading thread ID befor calling 

    /* set position to 1st frame of pick up thread # */
    if (setpos_1st_frame(vd) < 0) {
        ierr = -1;
    }
    return ierr;
}


int align_frame(vdif_file_t *vd, int numthread, int *frame)
{
/*
    align frame# of all thread to be the same
        2019-10-24

    [in]
       numthread  -- # of threads
    [return]
       frame   ---  frame# aligned
       atamadashi --  # difference between MAX frame# and MIN frame# before aligne

*/
    int i,n;
    int frame_min, frame_max;
    int sa,iw;

    frame_min = -1;
    frame_max = -1;

    for (n = 0; n < numthread; n++) {
        if (frame_min < 0) frame_min = vd[n].vhead.frame;
        if (frame_max < 0) frame_max = vd[n].vhead.frame;
        if(vd[n].vhead.frame < frame_min) frame_min = vd[n].vhead.frame;
        if(vd[n].vhead.frame > frame_max) frame_max = vd[n].vhead.frame;
    }
    sa = frame_min - frame_max;
    if (sa < 0) {
        iw = frame_min + vd[0].nframsec;
        frame_max = iw;
        frame_min = frame_max;
        sa=frame_min - frame_max;
    }
    if (sa > 0) { // adjust
        for (n = 0; n < numthread; n++) {
            iw=frame_max-vd[n].vhead.frame;
            if (iw > 0) {
                if(iw != vd[0].nframsec)
                    i = fwdNframeVDIF_multi(iw, &vd[n]);
            } 
        }
    }
    *frame = vd[0].vhead.frame;
    return sa;
}

void sort_table_thread_ascending(int *table_thread, int *table_frames, int nsize)
{
/*
    sorting table with ascending power
    2019-11-06
*/
    int n;
    int iw, ipos, np;
    int p1st;
    int *wt1, *wt2, *flag;

    wt1 = (int *)malloc(nsize * sizeof(int));
    wt2 = (int *)malloc(nsize * sizeof(int));
    flag = (int *)malloc(nsize * sizeof(int));
    for (n = 0; n < nsize; n++) {
        wt1[n] = table_thread[n];
        wt2[n] = table_frames[n];
        flag[n] = 1;
    }

    np = 0;
    while (1) {
        p1st = 1;
        for (n = 0; n < nsize; n++) {
            if (flag[n] == 1) {
                if (p1st == 1) {
                    iw = wt1[n];
                    ipos = n;
                    p1st = 0;
                }
                if (wt1[n] < iw) {
                    iw = wt1[n];
                    ipos = n;
                }
            }
        }
        table_thread[np] = wt1[ipos];
        table_frames[np] = wt2[ipos];
        flag[ipos] = 0;
        np++;
        if (np >= nsize) break;
    }

    free(wt1);
    free(wt2);
    free(flag);
}

int convVDIF2VDIF2(runp_t *runp)
/*
    VDIF to VDIF conversion 
        single-channel/thread to multi-channel/thead
        
        2019-10-23 
        2019-12-11 revised
*/
{
    int pcount,icount,opos;
    int iwpos, ibpos;
    int i,n,m;
    int iwrite;
    int ierr;
	int frame_c;   // total frame number counter
    int numch;
    //int numthread;
    int ns1sec;   // # of samples in 1 sec
    int ns1word;   // # of samples in 32bit-word
    int nword;
    int64_t total_frame; // # of frames read
    int64_t totalbyte;
    double total_sec;
    double secframe_in;  //  second per frame in VDIF in
    double secframe_out;  //  second per frame in VDIF out
    int nf;
    int fnum1, fnum2;
    //int table_thread[32], table_frames[32];
    //int table_thread_orig[32];
    int iok;
	//int seconds;
    //FILE *fo;
    FILE *fo;
    //vdif_file_t vd[32];     // for input   MAX 32 thread
    vdif_file_t *vd;     // for input
    vdif_file_t vd1st;
    //vdif_file_t *vdo;    // for output
    timeval_vdif_t time1st;
//    vdifconv_t ct;
	int *conv2bit;
    //int adbit;
	int conv2bit_1[]={0,1,2,3};   // offset binary (VDIF default)
    int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (Mark-5B default)
    int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer

    //vdif_header_t vh;  // for input
    vdif_header_t vho;  // for output    vho[nthread]
    u_int32_t head[8];      // for header portion [8]
    u_int32_t *obuf;        // for data field [nword]
    int64_t *obuf64;
    //int wd;
    vdifconv3_t ct;
    vdif_thread_info_t thd;  // 2019-11-12

    int sec_count;
    //fpos_t pos;  // for debug


    ierr=0;
	frame_c=0; 

	if(runp->e2bit==1){
	   //ct.conv2bit=conv2bit_1;
	   conv2bit=conv2bit_1;
	} else if(runp->e2bit==3){
	   //ct.conv2bit=conv2bit_3;
	   conv2bit=conv2bit_3;
	} else {
	   //ct.conv2bit=conv2bit_2;
	   conv2bit=conv2bit_2;
	}

    //if (i=get_num_thread_vdif(runp->vdifname, &numthread, table_thread, table_frames) < 0) {
    //if (i=get_num_thread_vdif(runp->vdifname, &thd) < 0) {
    if (i=get_num_thread_vdif(runp->vdifname, runp->skip_bytes,  &thd) < 0) { // 2019-12-11
        if (i == -1) {
            printf(" %s is not a VDIF file\n", runp->vdifname);
        }
        else {
            printf(" VDIF file (%s) open error\n",runp->vdifname);
        }
        goto exit2return;

    }

    vd = (vdif_file_t *)malloc(thd.numthread * sizeof(vdif_file_t));

    for (n = 0; n < thd.numthread; n++) vd[n].skip_bytes = runp->skip_bytes; // 2019-12-11

    // thread table sorting 
    memcpy(thd.tab_thread_orig, thd.tab_thread, 64 * sizeof(int));
    sort_table_thread_ascending(thd.tab_thread, thd.tab_frames, thd.numthread);

    if (runp->flag_ch_alloc == 1) {  // CH vs Thread# are set by operator
        for (n = 0; n < thd.numthread; n++) thd.tab_thread[n] = runp->chtable[n];
    }

    for (n = 0; n < thd.numthread; n++) {
        if (get_vdif_basic_info(&vd[n], runp) < 0) goto exit2return2;
    }
    for (n = 0; n < thd.numthread; n++) {
        if (ini_vdif_set_1st_frame(&vd[n], thd.tab_thread[n]) < 0) {
            printf(" Error at \"ini_vdif_set_1st_frame\"  for reading thread# %d\n", thd.tab_thread[n]);
            goto exit2return2;
        }
    }

    // check all thread have a same frame length and same # of channels and same AD bits
    iok = 1;
    for (n = 0; n < thd.numthread; n++) {
        //printf(" n=%d  frame_bytes=%d  numch=%d  adbit=%d\n",    // debug
        //    n, vd[n].frame_bytes, vd[n].numch, vd[n].adbit);     // debug
        if (vd[n].frame_bytes != vd[0].frame_bytes) iok = 0;
        if (vd[n].numch != vd[0].numch) iok = 0;
        if (vd[n].adbit != vd[0].adbit) iok = 0;
    }
    if (iok == 0) {
        printf(" Warning! Some threads have the different frame parameter!\n");
        printf(" Warning! All threads should have the same sampling and frame parameter!\n");
        goto exit2return;
    }

    disp_condition2(runp, vd, thd.numthread, thd.tab_thread_orig);
    disp_condition_ch_thread(runp, vd, thd.numthread, thd.tab_thread);


    // aligne all thread data with same frame #
    i = align_frame(vd, thd.numthread, &fnum1);
    printf(" djusted # of frames = %d,   aligned frame#= %d\n", i, fnum1);  // debug

    vd1st = vd[0];

//    disp_condition(runp, &vd);

    //if(vd.vhead.frame != 0){
/*
    if(runp->flag_1sec_adjust==1 && vd.vhead.frame != 0){
        printf(" Warning!! VDIF data don't start at on the second!!\n");
        printf("           so forward to the next on the second!!\n");
        //if(fwdNextSecondDIF(&vd) < 0) {
        if(fwdNsecVDIF_multi(1,&vd) < 0) {
            printf("  Error at forwarding!\n");
            exit(0);
        }
        
    }
*/	


    time1st=vd[0].vhead.t;  // save 1st time 
    //vd.pos_1st_header=vd.pos;  // save 1st header position to process 


    // assuming all threads are same frame length and sampling parameters

    for (n = 0; n < thd.numthread; n++) {
        secframe_in = (double)(vd[n].frame_bytes * 8 / (vd[n].numch*vd[n].adbit)) / vd[n].fsample;

        /*
          set reading buffer size to satisfy following condition
             1. (frame size)/(buffer size) should be an integer number
        */
        /*
        vd[n].nword = vd[n].frame_bytes / 4;
        while (vd[n].nword > BUFMAX) vd[n].nword /= 2;
        if ((vd[n].nword % 2) != 0) vd[n].nword *= 2;   // to treat 64-bit word 

        if (((vd[n].frame_bytes / 4) % vd[n].nword) != 0) {
            printf(" convVDIF2VDIF1: buffer size does\'nt satisfy the condition! %d %d\n", vd[n].nword, vd[n].frame_bytes * 4);
        }
        */

        /* set buffer size as same as the size of frame data portion */
        vd[n].nword = vd[n].frame_bytes / 4;
        vd[n].imax = (vd[n].frame_bytes / 4) / vd[n].nword;   // # of loops to read entire one frame data
        // now imax is always 1
        vd[n].nsample = vd[n].nword * 32 / vd[n].adbit;  // # of samples in buffer (don't care about # of Chs)

        vd[n].ubuf = (u_int32_t *)malloc(vd[n].nword * sizeof(u_int32_t)); //  data read buffer
        //ubuf64 = (int64_t *)vd.ubuf;

        vd[n].imax_c = 0;

        // debug
        //printf(" %d %d %d\n", n, vd[n].imax, vd[n].nsample);
    }
    
    /* foward startoffset seconds */
//	if(runp->soffset>0){
//		printf("  %d second forwading.....",runp->soffset);
//	    if(fwdNsecVDIF_multi(runp->soffset, &vd) <0){
//           printf("Foward error %d\n",runp->soffset);
//           i=fclose( vd.fin ) ;
//           vd.flag_open = 0;
//           goto exit2return;
//        }
//	    printf("\n");
//	    /* get headder info again */
//        if(vdif_header_read_mode1(&vd)<0) goto exit2return;
//	    time1st=vd.vhead.t;
//	}
//    vd1st = vd;

    /* preparation VDIF out */
    numch = thd.numthread;    // # of channels (= # of threads input VDIF)
    //numthread = 1;

    /* set output frame size to satisfy following condition
          1sec/1frame_period should be an integer number  */
    ns1sec = (int)vd[0].fsample; 
    ns1word = 32 / (vd[0].adbit*numch);
    nword = DFLT_FRAME_SIZE;
    if ((ns1sec % (nword*ns1word)) != 0) {
        printf(" convVDIF2VDIF2: output frame size (%d 32-bit word) does\'nt satisfy the condition!\n", 
            nword);
        goto exit2return;
    }
    if ((nword % numch) != 0) {
        printf(" convVDIF2VDIF2: output frame size (%d 32-bit word) does\'nt satisfy the 2nd condition!\n", 
            nword);
        goto exit2return;
    }

    /* VDIF out header common info setup */

    //vho = (vdif_header_t *)malloc(numthread*sizeof(vdif_header_t));
    //head = malloc(sizeof(u_int32_t *) * numthread);
    //for (n = 0; n < numthread; n++) head[n] = malloc(8 * sizeof(u_int32_t));
    obuf = malloc(nword*sizeof(u_int32_t));
    //for (n = 0; n < numthread; n++) obuf[n] = malloc(nword * sizeof(u_int32_t));
    obuf64 = (int64_t *)obuf;

    /* VDIF out header common info setup */
    ct.vdifadbit = vd1st.adbit;
    ct.o_adbit = vd1st.adbit;
    ct.bs_vdif = vd1st.adbit*numch;
    ct.numthread = 1;
    ct.o_numch = thd.numthread;
    ct.i_numthread = thd.numthread;

    ct.nframesec = ns1sec / ((nword * 32) / ct.bs_vdif);  // check
    secframe_out = (double)(nword * 32 / ct.bs_vdif)/vd[0].fsample;   // sec per frame 

    // get frame number in VDIF out
    get_frame_num(secframe_in, vd1st.vhead.frame, secframe_out, &fnum1, &fnum2);

    if (vd1st.vhead.frame > 0 ){
        if (fnum1 == 0) {  // adjust on the second is necessary
            printf(" adjusting on the second is required!\n");
            for(n=0; n<thd.numthread; n++){
                if (fwdNsecVDIF_multi(1, &vd[n]) < 0) {
                    printf("  Error at forwarding!\n");
                    exit(0);
                }
            }
        }
        else {
            nf = fnum1 - vd1st.vhead.frame;
            if (nf > 0) {  // proceed this frame difference
                printf(" forwarding %d frames is required for input VDIF file!\n", n);
                for (n = 0; n < thd.numthread; n++) {
                    if (fwdNframeVDIF_multi(nf, &vd[n]) < 0) {
                        printf("  Error at frame forwarding!\n");
                        exit(0);
                    }
                }
            }
        }
    } 

    // debug
//    for (n = 0; n < numthread; n++) {
//        printf("%d pos %lld  thread# %d %d  frame# %d\n",n, vd[n].pos,
//            vd[n].vhead.thread_id, vd[n].pup_thread,vd[n].vhead.frame);
//    }

    /* get headder info again */
    for (n = 0; n < thd.numthread; n++) {
        if (vdif_header_read_mode1_multi(&vd[n]) < 0) goto exit2return;
//        for (i = 0; i < numthread; i++) {
//            fgetpos(vd[i].fin, &pos);
//            printf(" %lld", pos);
//        }
//        printf("\n");
    }
    // debug
//    printf("\n\n");
//    for (n = 0; n < numthread; n++) {
//        i = fgetpos(vd[n].fin, &pos);  
//        printf("%d pos %lld (%lld)  thread# %d %d  frame# %d\n", n, vd[n].pos,pos,
//            vd[n].vhead.thread_id, vd[n].pup_thread, vd[n].vhead.frame);
//    }
    time1st = vd[0].vhead.t;
    vd1st = vd[0];

    vho = vd1st.vhead;
    vho.frame = fnum2; // usually unequal!!

    /* output file create */
    if(runp->oname[0]=='\0'){
       oname_make(runp);
    }

    //printf("  Debug fnum1, fnum2  %d  %d\n", fnum1, fnum2);  //  debug
    printf("\n");
    printf(" ======================= INFORMATION FOR OUTPUT VDIF FILE =====================\n");
    printf("  File Name    : %s\n", runp->oname);
    printf("  Start Time   : %4d/%02d/%02d %02d:%02d:%02d        Frame#  : %d\n",
        time1st.year, time1st.month, time1st.mday, time1st.hh, time1st.mm, time1st.ss,
        fnum2);
    printf("  # of Threads : %d                          Thread# : %d\n", ct.numthread, runp->set_tid);
    printf("  # of CHs     : %d\n", ct.o_numch);
    //printf("  Start Frame# : %d\n", fnum2);
    printf(" ==============================================================================\n");
    printf("\n");


    //printf(" VDIF Data File will be created : %s\n",runp->oname);

    if( (fo  = fopen( runp->oname, "wb" )) == NULL ){
           printf("      File  %s  Open Error! \n",runp->oname);
           ierr= -2;
           goto exit2return;
    }


    /* VDIF initial header create */
    make_vdif_header2(&vd1st, head, 4 * nword+32, runp->set_tid, numch);
    vho.sec_fm_ref = vd1st.vhead.sec_fm_ref;
    update_vdif_header2(vho.sec_fm_ref, vho.frame, head);

    pcount=0; // converted data period counter (in sec)

    ct.nbit=0;
    icount=0;  // 32bit word counter
    opos=0;    // out buffer writing position

    iwpos = 0;
    ibpos = 0;

    sec_count = 0;

    total_frame = 0;
    totalbyte=0;
    total_sec = 0.0;

    /* header monitor */
    printf(" ======================= INFORMATION FOR INPUT VDIF FILE ======================\n");
	printf("                                                             EXTENDED DATA\n");
	printf(" HH:MM:SS  FRAME#  I  L  VER  CHS     #Bytes  C AD  TID SID       W1       W2\n");
    printf(" ------------------------------------------------------------------------------\n");


    while(1){
        /* VDIF 1frame read */

        for (n = 0; n < thd.numthread; n++) {
            i = vdif_read_frame_multi(&vd[n]);
            if (i < 0) goto exithere;

            if(n==0) total_frame++;     // count frame # for 1st thread# only 
            //if (n == 0) {// debug
            //    printf(" Total_frame  %d\n", total_frame);
            //    printf(" %02d:%02d:%02d %6d", vd[0].vhead.t.hh, vd[0].vhead.t.mm, vd[0].vhead.t.ss, vd[0].vhead.frame);
            //    printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %08X %08X\n",
            //        vd[0].vhead.invalid, vd[0].vhead.legacy, vd[0].vhead.ver,
            //        (int)pow(2.0, vd[0].vhead.log2_chs), vd[0].vhead.frame_len64 * 8,
            //        vd[0].vhead.d_type, vd[0].vhead.bpsampl + 1, vd[0].vhead.thread_id, vd[0].vhead.sid, vd[0].vhead.w1, vd[0].vhead.w2);
            //}
        }

        if (vd[0].vhead.frame == 0) {

            sec_count++;

            printf(" %02d:%02d:%02d %6d", vd[0].vhead.t.hh, vd[0].vhead.t.mm, vd[0].vhead.t.ss, vd[0].vhead.frame);
            printf(" %3d %2d  %3d   %2d %10d  %1d %2d %4d  %s %08X %08X\n",
                vd[0].vhead.invalid, vd[0].vhead.legacy, vd[0].vhead.ver,
                (int)pow(2.0, vd[0].vhead.log2_chs), vd[0].vhead.frame_len64 * 8,
                vd[0].vhead.d_type, vd[0].vhead.bpsampl + 1, vd[0].vhead.thread_id, vd[0].vhead.sid, vd[0].vhead.w1, vd[0].vhead.w2);
            
            // make next frame header
            vho.sec_fm_ref = vd[0].vhead.sec_fm_ref;
            vho.frame = 0;
            update_vdif_header2(vho.sec_fm_ref, vho.frame, head);

        }

        // make output frame data
        
        // convert to VDIF multi-channel data (max 32bit word)
        //ct.owVDIF = obuf[opos];   // now # of threads of output VDIF is 1
        //ct.owVDIF64 = obuf64[opos];   // now # of threads of output VDIF is 1
        //conv_vdifmch2s(&ct);
        for (m = 0; m < vd[0].nword; m++) {
            for (n = 0; n < thd.numthread; n++) {
                ct.iwVDIF[n] = vd[n].ubuf[m];
            }
            ct.owVDIF = obuf + opos;
            conv_vdifsch2m(&ct);
            //obuf64[opos] = ct.owVDIF64;
            //if (ct.nbit == 64) {
            //    ct.nbit = 0;
            //    opos++;
            //}
            opos += thd.numthread;
            if (opos >= nword) {
                iwrite = fwrite(head, sizeof(int), 8, fo);
                totalbyte += iwrite * sizeof(int);
                iwrite = fwrite(obuf, sizeof(int), nword, fo);
                totalbyte += iwrite * sizeof(int);
                vho.frame++;
                // make next frame header
                update_vdif_header2(vho.sec_fm_ref, vho.frame, head);
                total_sec += secframe_out;
                opos = 0;
            }
        }

        // period check
        if(runp->period > 0){
            if(sec_count >= runp->period) goto exithere;
        }
	}
exithere:
    for (n = 0; n < thd.numthread; n++) {
        if (vd[n].flag_open == 1) i = fclose(vd[n].fin);
        vd[n].flag_open = 0;
    }
    // rest of data write

    if(opos != 0){
       //printf("convVDIF2VDIF2:  write final frame.\n");
        iwrite = fwrite(head, sizeof(int), 8, fo);
        totalbyte += iwrite * sizeof(int);
        iwrite = fwrite(obuf, sizeof(int), nword, fo);
        totalbyte += iwrite * sizeof(int);
    }
    if(totalbyte==0) printf("  NO DATA WERE CONVERTED!!\n");
    printf(" ==============================================================================\n");
	printf("\n");
    i=fclose( fo ) ;



exit2return:
       //printf("  K5 file (%s) has been created.\n\n",runp->k5name);
    for(n=0; n<thd.numthread; n++) free(vd[n].ubuf);
    //free(vho);
    //for (n = 0; n < numthread; n++) free(head[n]);
    //free(head);
    //for (n = 0; n < numthread; n++) free(obuf[n]);
    free(obuf);

    printf("\n");
//    if(ct.k5nfile==1){
    printf(" VDIF Data File created : %s\n",runp->oname);
    if(totalbyte ==0){
        printf("                             but contents are empty!!\n");
    } else {
        printf("                            (%lld bytes)\n",(long long)totalbyte);
        printf("                            (%f seconds)\n",total_sec);
    }

exit2return2:

    for (n = 0; n < thd.numthread; n++) {
        if (vd[n].flag_open == 1) i = fclose(vd[n].fin);
        vd[n].flag_open = 0;
    }

    return 0;
}


void oname_make(runp_t *rp)
{
/*
  Create VDIF out file name  
         2019-10-23
*/
    int i,len;
    //int flag_remove=0;
    int flag_remove;
    char vdifdir[256],dir[256], name[256],fname[256]; 
    char dir2[256],name2[256];

    flag_remove=0;
    if(rp->naming_type==1) flag_remove=1;

    //parse_dir_and_name(adsname, dir, name);
    parse_dir_and_name(rp->vdifname, dir, name);

    /* remove file extension if set so */

    if(flag_remove==1){
        len=strlen(name);

         // find last "." in the name to remove extension
        for(i=len-1; i>=0; i--){
           if(name[i]=='.'){
               name[i]='\0';
               break;
           }
        }
    }


    // out directory prepare
    if(rp->outdir[0] == '\0'){
        vdifdir[0]='\0';
    } else {
        // converting '\' to '/' if so
        len=strlen(rp->outdir);
        //printf("C %d %s\n",len,outdir);
        for(i=0; i<len; i++){
            if(rp->outdir[i]=='\\'){
                vdifdir[i]='/';
            } else {
                vdifdir[i]=rp->outdir[i];
            }
        }
        vdifdir[len]='\0';
        // checkdir name accompanied by "/" or not 
        if(vdifdir[len-1] != '/') {
           vdifdir[len]='/';
           vdifdir[len+1]='\0';
        }
    }

    // check given oname is accompanied by directory or not
    dir2[0]='\0'; name2[0]='\0';
    if(rp->oname_given[0] != '\0'){
        parse_dir_and_name(rp->oname_given, dir2, name2);
    }

    // create oname

    if(rp->oname_given[0] == '\0'){  // in case not given from an operator
        strcpy(fname,name);
        if (rp->naming_type == 0) {
            strcat(fname, ".mch");
        }
        else {
            strcat(fname, ".mch.vdif");
        }
        strcpy(rp->oname,vdifdir);
        strcat(rp->oname,fname);
    } else {   // in case of given from an operator
        if(dir2[0]=='\0'){
            if(rp->outdir[0]=='\0'){ 
                strcpy(rp->oname,rp->oname_given);
            } else {
                strcpy(rp->oname,vdifdir);
                strcat(rp->oname,name2);
            }
        } else {
            strcpy(rp->oname,rp->oname_given);
        }
    }
}


void make_vdif_header2(vdif_file_t *vd, u_int32_t *head, int frame_byte, int nthread, int numch)
/*
     VDIF header block creation
     2019-10-21 

  [in]
      vd     --  vdif_file_t type variable
      nthread --  thread #
  [out]
      head  ---- pointer to header array (8 X 32 bits) buffer 
*/
{
    u_int32_t w1,wd;
    int nframe64;
    int log2ch;

    // copy header

    memcpy(head, vd->head, 4 * 8);

    // set # of CHs 
    log2ch = (int)((log10((double)numch) / log10(2.0)) + 0.5);
    w1 = (log2ch & 0x1F) << 24;
    wd = head[2];
    wd = ((wd & 0xE0FFFFFF) | w1);   // set CH field (bit# 24-28)   (log2(CHS))
    //head[2] = wd;

    // set frame length
    nframe64 = frame_byte / 8;
    w1 = (nframe64 & 0xFFF);
    wd = ((wd & 0xFFFFF000) | w1);   // set thread# field (bit# 0-23)
    head[2] = wd;

    // set thread #
    wd = head[3];
    w1 = (nthread & 0x3FF) << 16;
    wd = ((wd & 0xFC00FFFF) | w1);   // set thread# field (bit# 16-25)
    head[3] = wd;    
}


void update_vdif_header2(int secfmepoch, int frame, u_int32_t *head)
/*
     update VDIF header block for time, frame #
     2019-10-30

  [in]
      secfmepoch --- seconds from epoch
      frame      --- frame #
      nthread    --- thread #
  [out]
      head  ---- pointer to header array (8 X 32 bits) buffer
*/
{

    u_int32_t w1, wd;

    // set seconds from reference epoch field
    wd = head[0];
    w1 = (secfmepoch & 0x3FFFFFFF);
    wd = ((wd & 0xC0000000) | w1);  // set bit# 0-29
    head[0] = wd;

    // set frame number
    wd = head[1];
    w1 = (frame & 0xFFFFFF);
    wd = ((wd & 0xFF000000) | w1);  // set bit# 0-23
    head[1] = wd;

    // set thread #
//    wd = head[3];
//    w1 = (nthread & 0x3FF) << 16;
//    wd = ((wd & 0xFC00FFFF) | w1);   // set thread# field (bit# 16-25)
//    head[3] = wd;
}


/* crc subroutine ripoff  by David*/
int crcc(unsigned char* idata, int len, int mask, int cycl) {
	unsigned int istate = 0;
	int idbit;
	int q, ich, icb;

	for (idbit = 1; idbit <= len; idbit++) {
		q = istate & 1;
		ich = (idbit - 1) / 8;
		icb = 7 - (idbit - 1) % 8;
		if ((((idata[ich] >> icb) & 1) ^ q) == 0)
			istate &= -2;
		else {
			istate ^= mask;
			istate |= 1;
		}
		istate = (istate >> 1) | ((istate & 1) << (cycl -1));  
	}

	return ((int) istate);
}


void m5b_time_update(M5B_HEADER_T *m5b, int mjd1st, int seconds_init, int sec_count)
{
   // time update for Mark5B 
    double dsec,dmsec10;
    dsec=(double)m5b->frame*m5b->p_frame;   // fractional second
    dmsec10=dsec*10000.0;
    m5b->t.msec10=(int)dmsec10;
    m5b->t.second=seconds_init+sec_count;
    m5b->t.mjd=mjd1st+(int)((double)m5b->t.second/86400.0);
    if(m5b->t.second >= 86400) m5b->t.second-=86400;
}


void conv_vdifsch2m(vdifconv3_t *ct)
/*
    VDIF(single-channel) to VDIF(multi-channel) conversion engine 
    2019-10-25

       ***  2bit conversion table is fixed) ****

        VDIF format is fixed to as follows


  [in]
     iwVDIF[32]       --- VDIF  32bit word data  (max 32 thread)
     o_adbit     ---   AD bits  (1 or 2) 
     nbit        --- already filled out bits in owVDIF64 (starting at 0 and <64)
     i_numthread   --- # of thread in input VDIF

  [return]
     nbit        --- already filled out bits in owVDIF64 (maximum is 64)
     owVDIF64     --- VDIF 64bit word data

*/
{
    int n;
    //int iw, ik5, nb, ipick;
    int iw, ow;
    int nb, ipick;
    int bpos;  // bit position in a word
    //int64_t ow64;
    int *ow32;
    int ipos;
    //int opos;
    //int vdifch;
    //int64_t mask;
    int mask;
    int imask[] = { 0xFFFFFFFE,0xFFFFFFFD,0xFFFFFFFB,0xFFFFFFF7,
                 0xFFFFFFEF,0xFFFFFFDF,0xFFFFFFBF,0xFFFFFF7F,
                 0xFFFFFEFF,0xFFFFFDFF,0xFFFFFBFF,0xFFFFF7FF,
                 0xFFFFEFFF,0xFFFFDFFF,0xFFFFBFFF,0xFFFF7FFF,
                 0xFFFEFFFF,0xFFFDFFFF,0xFFFBFFFF,0xFFF7FFFF,
                 0xFFEFFFFF,0xFFDFFFFF,0xFFBFFFFF,0xFF7FFFFF,
                 0xFEFFFFFF,0xFDFFFFFF,0xFBFFFFFF,0xF7FFFFFF,
                 0xEFFFFFFF,0xDFFFFFFF,0xBFFFFFFF,0x7FFFFFFF };
    int ibset[] = { 0x00000001,0x00000002,0x00000004,0x00000008,
                 0x00000010,0x00000020,0x00000040,0x00000080,
                 0x00000100,0x00000200,0x00000400,0x00000800,
                 0x00001000,0x00002000,0x00004000,0x00008000,
                 0x00010000,0x00020000,0x00040000,0x00080000,
                 0x00100000,0x00200000,0x00400000,0x00800000,
                 0x01000000,0x02000000,0x04000000,0x08000000,
                 0x10000000,0x20000000,0x40000000,0x80000000 };

    int imask2[] = { 0xFFFFFFFC,0xFFFFFFF3,0xFFFFFFCF,0xFFFFFF3F,
                  0xFFFFFCFF,0xFFFFF3FF,0xFFFFCFFF,0xFFFF3FFF,
                  0xFFFCFFFF,0xFFF3FFFF,0xFFCFFFFF,0xFF3FFFFF,
                  0xFCFFFFFF,0xF3FFFFFF,0xCFFFFFFF,0x3FFFFFFF };

    int imask4[] = { 0xFFFFFFF0,0xFFFFFF0F,0xFFFFF0FF,0xFFFF0FFF,
                  0xFFF0FFFF,0xFF0FFFFF,0xF0FFFFFF,0x0FFFFFFF };

    int imask8[] = { 0xFFFFFF00,0xFFFF00FF,0xFF00FFFF,0x00FFFFFF };

    ow32 = ct->owVDIF;

    /* set mask bits */

    mask = 0x1;
    if (ct->vdifadbit == 2) mask = 0x3;
    if (ct->vdifadbit == 4) mask = 0xF;
    if (ct->vdifadbit == 8) mask = 0xFF;


    nb = 0;
    while(1){
        for (n = 0; n < ct->i_numthread; n++) {
            iw = ct->iwVDIF[n];
            ipick = (int)(iw & mask);
            ipos = nb / 32;
            bpos = nb % 32;
            ow = ow32[ipos];
            // put data in VDIF out
            if (ct->o_adbit == 8) {  // convert to 8bit mode
                ow = (ow & imask8[bpos / 8]) | (ipick << bpos);
            }
            else if (ct->o_adbit == 4) {  // convert to 4bit mode
                ow = (ow & imask4[bpos / 4]) | (ipick << bpos);
            }
            else if (ct->o_adbit == 2) {  // convert to 2bit mode
            //ipick=conv2bit[ipick];
                ow = (ow & imask2[bpos / 2]) | (ipick << bpos);
            }
            else {    // pick up 1bit data
               //ipick >>= 1;   // MSB get
                ow = (ow & imask[bpos]) | (ipick << bpos);
            }
            ow32[ipos] = ow;
            //nb += ct->o_adbit;
            iw >>= ct->vdifadbit;
            ct->iwVDIF[n]=iw;
            nb += ct->o_adbit;

        }
        if (nb >= 32 * ct->i_numthread) break;
    }

}

int vdif_read_frame_multi(vdif_file_t *vd)
{
    /*
         VDIF data 1 frame read for multi thread case
         2019-10-31

         File pointer is set to the start of next frame (can be different thread#)

         [in]
            vd->pup_thread  -- target thread#

         [return]
            -VE  -- error occurred


    */
    int i;

    i = vdif_header_read_mode2_multi(vd);
    if (i < 0) return -1;

    vd->wordread = fread(vd->ubuf, sizeof(unsigned int), vd->nword, vd->fin);
    if (vd->wordread == 0) { // EOF
        i = fclose(vd->fin);
        vd->flag_open = 0;
        return -1;
    }
    //vd->pos += vd->frame_bytes;   // use fgetpos instead
    //fgetpos(vd->fin, &vd->pos);
    i = fgetpos(vd->fin, (fpos_t *)&vd->pos);  // 2019-11-01 (to avoid warning at linux)
    return 0;
}